import re
import logging
from . import common
from ..enum import ProtectionLevel, CollaboratorType, SearchType, PortalAccountType, FileAccessMode
from ..types import PortalAccount, UserAccount, GroupAccount
from ...common import Object
from ...exception import CTERAException


def _search_collaboration_member(ctera_host, account, cloud_folder_uid):
    search_param = Object()
    if account.account_type == PortalAccountType.User:
        search_param.searchType = SearchType.Users
    elif account.account_type == PortalAccountType.Group:
        search_param.searchType = SearchType.Groups
    else:
        raise CTERAException("Invalid account type", None, account_type=account.account_type)

    search_param.searchTerm = account.name
    search_param.resourceUid = cloud_folder_uid
    search_param.countLimit = 1
    response = ctera_host.execute('', 'searchCollaborationMembers', search_param)
    if not response.objects:
        logging.getLogger().warning('Could not find results that match your search criteria. %s', {'name': account.name})
    elif len(response.objects) > 1:
        logging.getLogger().warning('Search returned multiple results. Skipping...')
    else:
        collaborator = response.objects.pop()
        if account.is_local:
            if account.account_type == PortalAccountType.User and collaborator.type == CollaboratorType.LU:
                logging.getLogger().debug('Found local user account. %s', {'name': account.name})
                return collaborator
            if account.account_type == PortalAccountType.Group and collaborator.type == CollaboratorType.LG:
                logging.getLogger().debug('Found local group account. %s', {'name': account.name})
                return collaborator
            logging.getLogger().warning('Expected to find local %s but found domain %s', account.account_type, account.account_type)
        else:
            if account.account_type == PortalAccountType.User and collaborator.type == CollaboratorType.DU:
                logging.getLogger().debug('Found domain user account. %s', {'name': account.name, 'domain': account.directory})
                return collaborator
            if account.account_type == PortalAccountType.Group and collaborator.type == CollaboratorType.DG:
                logging.getLogger().debug('Found domain group account. %s', {'name': account.name, 'domain': account.directory})
                return collaborator
            logging.getLogger().warning('Expected to find domain %s but found local %s', account.account_type, account.account_type)
    return None


def remove_share_recipients(ctera_host, path, accounts):
    share_info = get_share_info(ctera_host, path)
    current_accounts = _obtain_current_accounts(share_info)

    accounts_to_keep, accounts_to_remove = [], []
    if current_accounts:
        for idx, current_account in enumerate(current_accounts):
            if current_account not in accounts:
                accounts_to_keep.append(share_info.shares[idx])
            else:
                logging.getLogger().debug('Found recipient to remove. %s', {'account': str(current_account)})
                accounts_to_remove.append(current_account)
        if not accounts_to_remove:
            logging.getLogger().info(
                'Could not find share recipients to remove. %s',
                {
                    'path': str(path.relativepath),
                    'recipients': [str(member) for member in accounts]
                }
            )
        else:
            share_param = _create_share_param(path.fullpath(), share_info.teamProject, share_info.allowReshare, share_info.shouldSync)
            _update_share_param(share_param, accounts_to_keep)
            logging.getLogger().info(
                'Removing share recipients. %s',
                {'path': str(path.relativepath), 'recipients': [str(account) for account in accounts_to_remove]}
            )
            ctera_host.execute('', 'shareResource', share_param)
    else:
        logging.getLogger().info('Share has no recipients. %s', {'path': str(path.relativepath)})
    return accounts_to_remove


def get_share_info(ctera_host, path):
    return ctera_host.execute('', 'listShares', path.encoded_fullpath())


def share(ctera_host, path, recipients, as_project, allow_reshare, allow_sync):

    if _is_subfolder(path):
        as_project = False  # Can't be a team project
        allow_sync = False  # Can't allow sync of a sub-folder

    valid_recipients = _obtain_valid_recipients(ctera_host, path, recipients)
    if valid_recipients:
        share_param = _create_share_param(path.fullpath(), as_project, allow_reshare, allow_sync)
        for recipient in valid_recipients:
            _add_share_recipient(share_param, recipient)

        logging.getLogger().info(
            'Sharing item. %s',
            {'path': str(path.relativepath), 'recipients': [str(recipient) for recipient in valid_recipients]}
        )
        ctera_host.execute('', 'shareResource', share_param)
        return valid_recipients
    logging.getLogger().warning('Resource was not shared. Could not find valid recipients. %s', {'path': str(path.relativepath)})
    return valid_recipients


def _is_sharing_on_user_behalf(path):
    return re.match(r'^Users/[^/]+(?=/)', path.relativepath.as_posix()) is not None


def _is_subfolder(path):
    depth = len(path.relativepath.parts)
    sharing_on_user_behalf = _is_sharing_on_user_behalf(path)
    return (sharing_on_user_behalf and depth > 3) or (not sharing_on_user_behalf and depth > 1)


def add_share_recipients(ctera_host, path, recipients):
    share_info = get_share_info(ctera_host, path)
    current_accounts = _obtain_current_accounts(share_info)
    valid_recipients = _obtain_valid_recipients(ctera_host, path, recipients)
    share_param = _create_share_param(path.fullpath(), share_info.teamProject, share_info.allowReshare, share_info.shouldSync)
    _update_share_param(share_param, share_info.shares)
    accounts_added = []
    if valid_recipients:
        for recipient in valid_recipients:
            if recipient.account not in current_accounts:
                logging.getLogger().debug('Found new share recipient. %s', {'recipient': str(recipient.account)})
                _add_share_recipient(share_param, recipient)
                accounts_added.append(recipient)
            else:
                logging.getLogger().debug('Share recipient already exists. %s', {'recipient': str(recipient.account)})
        if accounts_added:
            logging.getLogger().info(
                'Adding share recipients. %s',
                {'path': str(path.relativepath), 'recipients': [str(recipient.account) for recipient in accounts_added]}
            )
            ctera_host.execute('', 'shareResource', share_param)
        else:
            logging.getLogger().warning('Could not find new share recipients. %s', {'path': str(path.relativepath)})
        return accounts_added
    logging.getLogger().warning('Share recipients were not added. Could not find valid recipients. %s', {'path': str(path.relativepath)})
    return valid_recipients


def _obtain_current_accounts(share_info):
    current_accounts = []
    for collaborator in share_info.shares:
        if collaborator.invitee.type == CollaboratorType.EXT:
            current_accounts.append(collaborator.invitee.email)
        else:
            current_accounts.append(PortalAccount.from_collaborator(collaborator.invitee))
    return current_accounts


def _obtain_valid_recipients(ctera_host, path, recipients):
    resource_info = common.get_resource_info(ctera_host, path)
    cloud_folder_uid = resource_info.cloudFolderInfo.uid
    valid_recipients = []
    for recipient in recipients:
        if _valid_recipient(recipient):
            if not recipient.type == CollaboratorType.EXT:
                collaborator = _search_collaboration_member(ctera_host, recipient.account, cloud_folder_uid)
                if collaborator:
                    recipient.collaborator = collaborator
                    valid_recipients.append(recipient)
            else:
                valid_recipients.append(recipient)
    return valid_recipients


def unshare(ctera_host, path):
    resource_info = common.get_resource_info(ctera_host, path)
    as_project, allow_reshare, allow_sync = True, True, True

    if _is_subfolder(path):
        as_project = False
        allow_sync = False

    logging.getLogger().info(
        'Unsharing %s. %s',
        ('folder' if resource_info.isFolder else 'file'), {'path': str(path.relativepath)}
    )
    share_param = _create_share_param(path.fullpath(), as_project, allow_reshare, allow_sync)
    return ctera_host.execute('', 'shareResource', share_param)


def _valid_recipient(recipient):
    valid_recipient = True
    if not recipient.account:
        logging.getLogger().warning(
            'No account information found. Skipping. %s',
            {'account': recipient.account}
        )
        valid_recipient = False
    if not isinstance(recipient.account, (str, UserAccount, GroupAccount)):
        logging.getLogger().warning(
            'Invalid recipient type. Skipping. %s',
            {'type': type(recipient.account)}
        )
        valid_recipient = False
    if recipient.access not in [v for k, v in FileAccessMode.__dict__.items() if not k.startswith('_')]:
        logging.getLogger().warning(
            'Invalid file access mode. Skipping. %s',
            {'account': str(recipient.account), 'access': recipient.access}
        )
        valid_recipient = False
    if recipient.type == CollaboratorType.EXT and not isinstance(recipient.account, str):
        logging.getLogger().warning(
            'Invalid external recipient type. Skipping. %s',
            {'expected': 'str', 'actual': type(recipient.account)}
        )
        valid_recipient = False
    if recipient.type in [CollaboratorType.LU, CollaboratorType.LG] and not recipient.account.is_local:
        logging.getLogger().warning(
            'Expected local account. Received a domain account. Skipping. %s',
            {'account': str(recipient.account)}
        )
        valid_recipient = False
    if recipient.type in [CollaboratorType.DU, CollaboratorType.DG] and recipient.account.is_local:
        logging.getLogger().warning(
            'Expected domain account. Received a local account. Skipping. %s',
            {'account': str(recipient.account)}
        )
        valid_recipient = False
    return valid_recipient


def _create_share_param(url, as_project, allow_reshare, allow_sync):
    share_param = Object()
    share_param._classname = 'ShareResourceParam'  # pylint: disable=protected-access
    share_param.url = url
    share_param.teamProject = as_project
    share_param.allowReshare = allow_reshare
    share_param.shouldSync = allow_sync
    share_param.shares = []
    return share_param


def _update_share_param(share_param, shares):
    for recipient in shares:
        share_config_param = Object()
        share_config_param._classname = 'ShareConfig'  # pylint: disable=protected-access
        share_config_param.accessMode = recipient.accessMode
        share_config_param.expiration = recipient.expiration
        share_config_param.protectionLevel = recipient.protectionLevel
        share_config_param.invitee = recipient.invitee
        share_param.shares.append(share_config_param)
    return share_param


def _add_share_recipient(share_param, recipient):
    share_config_param = Object()
    share_config_param._classname = 'ShareConfig'  # pylint: disable=protected-access
    share_config_param.accessMode = recipient.access
    if not recipient.access == FileAccessMode.NA:
        share_config_param.expiration = recipient.expiration_date
    invitee = None
    if recipient.type == CollaboratorType.EXT:
        invitee = Object()
        invitee._classname = 'Collaborator'  # pylint: disable=protected-access
        invitee.type = recipient.type
        invitee.email = recipient.account
        if recipient.two_factor:
            share_config_param.protectionLevel = ProtectionLevel.Email
        else:
            share_config_param.protectionLevel = ProtectionLevel.Public
    else:
        invitee = recipient.collaborator
    share_config_param.invitee = invitee
    share_param.shares.append(share_config_param)



import atexit
import logging
import shutil
import tempfile

from .registry import Registry


class TempfileServices:

    __tempdir_prefix = 'cterasdk-'

    @staticmethod
    def mkdir():
        registry = Registry.instance()
        tempdir = registry.get('tempdir')
        if tempdir is None:
            logging.getLogger().debug('Creating temporary directory.')
            tempdir = tempfile.mkdtemp(prefix=TempfileServices.__tempdir_prefix)
            logging.getLogger().debug('Temporary directory created. %s', {'path': tempdir})
            registry.register('tempdir', tempdir)
        return tempdir

    @staticmethod
    def mkfile(prefix, suffix):
        tempdir = TempfileServices.mkdir()
        logging.getLogger().debug('Creating temporary file.')
        fd, filepath = tempfile.mkstemp(prefix=prefix, suffix=suffix, dir=tempdir)
        logging.getLogger().debug('Temporary file created. %s', {'path': filepath})
        return (fd, filepath)

    @staticmethod
    @atexit.register
    def rmdir():
        registry = Registry.instance()
        tempdir = registry.get('tempdir')
        if tempdir is not None:
            logging.getLogger().debug('Removing temporary directory. %s', {'path': tempdir})
            shutil.rmtree(path=tempdir)
            logging.getLogger().debug('Removed temporary directory. %s', {'path': tempdir})
            registry.remove('tempdir')



import logging

from .common import SrcDstParam, ActionResourcesParam


def undelete(ctera_host, path):
    return undelete_multi(ctera_host, *[path])


def undelete_multi(ctera_host, *paths):
    undelete_param = ActionResourcesParam.instance()
    for path in paths:
        logging.getLogger().info('Recovering item. %s', {'path': str(path.relativepath)})
        undelete_param.add(SrcDstParam.instance(src=path.fullpath()))

    return ctera_host.execute('', 'restoreResources', undelete_param)



import functools
import logging
import socket

from ..common import Object
from ..convert import tojsonstr
from ..exception import HostUnreachable
from .cteraclient import CTERAClient, MigrationClient, RESTClient
from ..exception import CTERAException


def authenticated(function):
    @functools.wraps(function)
    def check_authenticated_and_call(self, *args, **kwargs):
        if self._is_authenticated(function, *args, **kwargs):  # pylint: disable=protected-access
            return function(self, *args, **kwargs)
        logging.getLogger().error('Not logged in.')
        raise CTERAException('Not logged in')

    return check_authenticated_and_call


class NetworkHost:
    def __init__(self, host, port, https):
        self._host = host
        self._port = port or 443 if https else 80
        self._https = https

    @property
    def _omit_fields(self):
        return []

    def test_conn(self):
        logging.getLogger().debug('Testing connection. %s', {'host': self.host(), 'port': self.port()})
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        rc = None
        try:
            rc = sock.connect_ex((self._host, self._port))
        except socket.gaierror:
            logging.getLogger().debug('Host unreachable. %s', {'host': self.host(), 'port': self.port()})
            raise HostUnreachable(None, self._host, self._port, self.scheme().upper())

        if rc != 0:
            logging.getLogger().debug('Host unreachable. %s', {'host': self.host(), 'port': self.port()})
            raise HostUnreachable(None, self._host, self._port, self.scheme().upper())

        logging.getLogger().debug('Host is reachable. %s', {'host': self.host(), 'port': self.port()})

    def scheme(self):
        return 'http' + ("s" if self._https else '')

    def host(self):
        return self._host

    def port(self):
        return self._port

    def https(self):
        return self._https

    def baseurl(self):
        return 'http' + ("s" if self._https else '') + '://' + self._host + ':' + str(self._port)

    def __str__(self):
        x = Object()
        x.__dict__ = {k: v for k, v in self.__dict__.items() if not (k.startswith('_') or k in self._omit_fields)}
        return tojsonstr(x)


class MigrationHost(NetworkHost):

    def __init__(self, host, port, https, is_authenticated=None, http_client=None):
        super().__init__(host, port, https)

        def always_authenticated(self, function):  # pylint: disable=unused-argument
            return True
        self._is_authenticated = is_authenticated if is_authenticated else always_authenticated
        self._client = MigrationClient(http_client) if http_client else RESTClient()

    @staticmethod
    def from_ctera_host(ctera_host):
        """Create a RESTful host instance from an existing CTERA host instance"""
        return MigrationHost(ctera_host.host(), ctera_host.port(), ctera_host.https(),
                             ctera_host._is_authenticated, ctera_host._ctera_client.http_client)  # pylint: disable=protected-access

    @authenticated
    def login(self, path):
        return self._client.login(self.baseurl(), path)

    @authenticated
    def get(self, path, params=None):
        return self._client.get(self.baseurl(), path, params or {})

    @authenticated
    def put(self, path, value):
        return self._client.put(self.baseurl(), path, value)

    @authenticated
    def post(self, path, value):
        return self._client.post(self.baseurl(), path, value)

    @authenticated
    def delete(self, path):
        return self._client.delete(self.baseurl(), path)


class CTERAHost(NetworkHost):  # pylint: disable=too-many-public-methods

    def __init__(self, host, port, https):
        super().__init__(host, port, https)
        self._ctera_client = CTERAClient(self._session_id_key)
        self._ctera_migrate = MigrationHost.from_ctera_host(self)
        self._session = None

    @property
    def _omit_fields(self):
        return super()._omit_fields + [
            'login',
            'logout'
        ]

    @property
    def base_api_url(self):
        raise NotImplementedError("Implementing class must implement the base_api_url property")

    @property
    def base_file_url(self):
        raise NotImplementedError("Implementing class must implement the base_api_url property")

    @property
    def _login_object(self):
        raise NotImplementedError(
            "Implementing class must implement the login_object property by returning an object with login and logout methods"
        )

    @property
    def _session_id_key(self):
        raise NotImplementedError("Implementing class must implement the _session_id_key property")

    def _is_authenticated(self, function, *args, **kwargs):
        raise NotImplementedError("Implementing class must implement the _is_authenticated method")

    def login(self, username, password):
        """
        Log in

        :param str username: User name to log in
        :param str password: User password
        """
        self._login_object.login(username, password)
        self._session.start_local_session(self)

    def logout(self):
        """ Log out """
        if self._session.active:
            self._login_object.logout()
            self._session.terminate()

    def session(self):
        return self._session

    def register_session(self, session):
        self._session = session

    def default_class(self, name):
        return self.get('/defaults/' + name)

    @authenticated
    def get(self, path, params=None, use_file_url=False):
        """ Retrieve a schema object as a Python object. """
        return self._ctera_client.get(self.base_file_url if use_file_url else self.base_api_url, path, params or {})

    @authenticated
    def openfile(self, path, params=None, use_file_url=False):
        return self._ctera_client.download(self.base_file_url if use_file_url else self.base_api_url, path, params or {})

    @authenticated
    def download_zip(self, path, form_data, use_file_url=False):
        return self._ctera_client.download_zip(self.base_file_url if use_file_url else self.base_api_url, path, form_data)

    @authenticated
    def show(self, path, use_file_url=False):
        """ Print a schema object as a JSON string. """
        print(tojsonstr(self.get(path, params={}, use_file_url=use_file_url), no_log=False))

    @authenticated
    def get_multi(self, path, paths, use_file_url=False):
        """ Retrieve one or more schema objects as a Python object. """
        return self._ctera_client.get_multi(self.base_file_url if use_file_url else self.base_api_url, path, paths)

    @authenticated
    def show_multi(self, path, paths, use_file_url=False):
        """ Print one or more schema objects as a JSON string. """
        print(tojsonstr(self.get_multi(path, paths, use_file_url=use_file_url), no_log=False))

    @authenticated
    def put(self, path, value, use_file_url=False):
        """ Update a schema object or attribute. """
        response = self._ctera_client.put(self.base_file_url if use_file_url else self.base_api_url, path, value)
        logging.getLogger().debug('Configuration changed. %s', {'url': path, 'value': tojsonstr(value, pretty_print=False)})
        return response

    @authenticated
    def post(self, path, value, use_file_url=False):
        response = self._ctera_client.post(self.base_file_url if use_file_url else self.base_api_url, path, value)
        logging.getLogger().debug('Added. %s', {'url': path, 'value': tojsonstr(value, pretty_print=False)})
        return response

    def form_data(self, path, form_data, use_file_url=False):
        return self._ctera_client.form_data(self.base_file_url if use_file_url else self.base_api_url, path, form_data)

    @authenticated
    def db(self, path, name, param, use_file_url=False):
        response = self._ctera_client.db(self.base_file_url if use_file_url else self.base_api_url, path, name, param)
        logging.getLogger().debug(
            'Database method executed. %s',
            {'url': path, 'name': name, 'param': tojsonstr(param, pretty_print=False)}
        )
        return response

    @authenticated
    def execute(self, path, name, param=None, use_file_url=False):
        """ Execute a schema object method. """
        response = self._ctera_client.execute(self.base_file_url if use_file_url else self.base_api_url, path, name, param)
        logging.getLogger().debug(
            'User-defined method executed. %s',
            {'url': path, 'name': name, 'param': tojsonstr(param, pretty_print=False)}
        )
        return response

    @authenticated
    def add(self, path, param, use_file_url=False):
        """ Add a schema object. """
        return self.db(path, 'add', param, use_file_url=use_file_url)

    @authenticated
    def delete(self, path, use_file_url=False):
        """ Delete a schema object. """
        response = self._ctera_client.delete(self.base_file_url if use_file_url else self.base_api_url, path)
        logging.getLogger().debug('Deleted. %s', {'url': path})
        return response

    @authenticated
    def mkcol(self, path, use_file_url=False):
        return self._ctera_client.mkcol(self.base_file_url if use_file_url else self.base_api_url, path)

    @authenticated
    def copy(self, src, dest, overwrite, use_file_url=False):
        return self._ctera_client.copy(self.base_file_url if use_file_url else self.base_api_url, src, dest, overwrite)

    @authenticated
    def move(self, src, dest, overwrite, use_file_url=False):
        return self._ctera_client.move(self.base_file_url if use_file_url else self.base_api_url, src, dest, overwrite)

    @authenticated
    def multipart(self, path, form_data, use_file_url=False):
        return self._ctera_client.multipart(self.base_file_url if use_file_url else self.base_api_url, path, form_data)

    @authenticated
    def upload(self, path, form_data, use_file_url=False):
        return self._ctera_client.upload(self.base_file_url if use_file_url else self.base_api_url, path, form_data)

    @authenticated
    def get_session_id(self):
        """
        Get the id of the current session

        :return str: Current session id
        """
        return self._ctera_client.get_session_id()

    def set_session_id(self, session_id):
        """
        Start a session with the session id instead of logging in

        :param str session_id: Session id for the new session
        """
        self._ctera_client.set_session_id(session_id)
        self._session.start_local_session(self)

    def set_authorization_headers(self, headers):
        """
        Start a session using authorization headers id instead of logging in

        :param dict headers: the authorization headers, represented as a key-value str dict
        """
        self._ctera_client.set_authorization_headers(headers)
        self._session.local_auth = True  # pylint: disable=protected-access
        self._session.start_local_session(self)

    def whoami(self):
        """
        Return the name of the logged in user.

        :return cterasdk.common.object.Object: The session object of the current user
        """
        return self._session.whoami()



from ..lib import FileSystem
from ..exception import CTERAException
from .base_command import BaseCommand


class UploadTaskStatus():
    IN_PROGRESS = 0
    COMPLETE = 1
    FAIL = -1


class Firmware(BaseCommand):
    """ Gateway Firmware upgrade API """

    def __init__(self, gateway):
        super().__init__(gateway)
        self._filesystem = FileSystem.instance()

    def upgrade(self, file_path, reboot=True, wait_for_reboot=True):
        """
        Upgrade the Filer firmware with the provided file

        :param str file_path: Path to the local file to upload
        :param bool,optional reboot: Perform reboot after uploading the new firmware, defaults to True
        :param bool,optional wait_for_reboot: Wait for reboot to complete (if reboot is performed), defaults to True
        """
        upload_task_info = self._upload_firmware(file_path)
        if upload_task_info.rc != 0:
            raise CTERAException(message='Failed to upload the new firmware', path=file_path)
        self._wait_for_completion(upload_task_info.taskPointer)
        if reboot:
            self._gateway.power.reboot(wait=wait_for_reboot)

    def _upload_firmware(self, file_path):
        file_info = self._filesystem.get_local_file_info(file_path)
        with open(file_path, 'rb') as fd:
            return self._gateway.upload(
                'proc/firmware',
                dict(
                    name='upload',
                    firmware=(file_info['name'], fd, file_info['mimetype'][0])
                )
            )

    def _wait_for_completion(self, task_pointer):
        while True:
            task_status = self._gateway.get(task_pointer)
            is_running = task_status.status == UploadTaskStatus.IN_PROGRESS
            if not is_running:
                if task_status.status == UploadTaskStatus.COMPLETE:
                    return
                raise CTERAException(
                    message=f'Filer failed to receive the new firmware - {task_status.statusMessage}',
                    instance=task_status
                )



# pylint: disable=wrong-import-position
from . import config

config.Logging.get()

from .common import Object, PolicyRule  # noqa: E402, F401
from .convert import fromjsonstr, tojsonstr, fromxmlstr, toxmlstr  # noqa: E402, F401
from .exception import CTERAException  # noqa: E402, F401
from .object import GlobalAdmin, ServicesPortal, Gateway, Agent  # noqa: E402, F401
from .core import query  # noqa: E402, F401
from .edge import types as gateway_types  # noqa: E402, F401
from .edge import enum as gateway_enum  # noqa: E402, F401
from .core import types as portal_types  # noqa: E402, F401
from .core import enum as portal_enum  # noqa: E402, F401
from .common import types as common_types  # noqa: E402, F401
from .common import enum as common_enum  # noqa: E402, F401



from ..client import NetworkHost, CTERAHost
from ..edge import backup
from ..edge import cli
from ..edge import logs
from ..edge import services
from ..edge import session
from ..edge import support
from ..edge import sync
from ..edge import uri


class Agent(CTERAHost):  # pylint: disable=too-many-instance-attributes
    """
    Main class operating on a Agent

    :ivar cterasdk.edge.backup.Backup backup: Object holding the Agent Backup APIs
    :ivar cterasdk.edge.cli.CLI cli: Object holding the Agent CLI APIs
    :ivar cterasdk.edge.logs.Logs logs: Object holding the Agent Logs APIs
    :ivar cterasdk.edge.services.Services services: Object holding the Agent Services APIs
    :ivar cterasdk.edge.support.Support support: Object holding the Agent Support APIs
    :ivar cterasdk.edge.sync.Sync sync: Object holding the Agent Sync APIs
    """

    def __init__(self, host, port=80, https=False, Portal=None):
        """
        :param str host: The fully qualified domain name, hostname or an IPv4 address of the Gateway
        :param int,optional port: Set a custom port number (0 - 65535), defaults to 80
        :param bool,optional https: Set to True to require HTTPS, defaults to False
        :param cterasdk.object.Portal.Portal,optional Portal: The portal throught which the remote session was created, defaults to None
        """
        super().__init__(host, port, https)
        self._remote_access = False
        self._session = session.Session(self.host())
        if Portal is not None:
            self._Portal = Portal
            self._ctera_client = Portal._ctera_client
            self._session.start_remote_session(self._Portal.session())
        self.backup = backup.Backup(self)
        self.cli = cli.CLI(self)
        self.logs = logs.Logs(self)
        self.services = services.Services(self)
        self.support = support.Support(self)
        self.sync = sync.Sync(self)

    @property
    def base_api_url(self):
        return uri.api(self)

    @property
    def _login_object(self):
        raise NotImplementedError("Currently login is not supported for Agent")

    @property
    def _session_id_key(self):
        return ''

    @property
    def _omit_fields(self):
        return super()._omit_fields + [
            'backup',
            'cli',
            'logs',
            'services',
            'support',
            'sync'
        ]

    def _is_authenticated(self, function, *args, **kwargs):
        return True

    def _api(self):
        return uri.api(self)

    def _baseurl(self):
        return NetworkHost.baseurl(self)



import urllib.parse
import logging
import time

import requests
import requests.exceptions as requests_exceptions
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

# from .ssl import CertificateServices
from ..convert import fromxmlstr, ParseException
from ..common import Object, merge
from .. import config
from ..exception import SSLException, HostUnreachable, ExhaustedException
from ..lib import ask


class HTTPException(Exception):

    def __init__(self, http_error):
        super().__init__()
        self.response = Object()
        self.response.code = http_error.response.status_code
        self.response.reason = http_error.response.reason
        try:
            self.response.body = fromxmlstr(http_error.response.text)
        except ParseException:
            self.response.body = ''
        if config.http['verbose']:
            self.response.headers = http_error.response.headers
            self.request = HTTPException._parse_request(http_error.request)

    @staticmethod
    def _parse_request(request):
        o = urllib.parse.urlparse(request.url)
        target = Object()
        target.host = Object()
        target.host.scheme = o.scheme
        target.host.hostname = o.hostname
        target.host.port = o.port
        target.method = request.method
        target.uri = o.path
        target.headers = request.headers
        if request.body is not None:
            target.body = request.body.decode('utf-8')  # decode from 'Bytes' to 'UTF-8'
        return target


class HTTPResponse:

    def __init__(self, response):
        self.url = response.url
        self.code = response.status_code
        self.headers = response.headers
        self.text = response.text

    def geturl(self):
        return self.url

    def getcode(self):
        return self.code

    def read(self):
        return self.text


def geturi(baseurl, path):
    path = urllib.parse.quote(path)
    count = int(baseurl.endswith('/')) + int(path.startswith('/'))
    if count == 2:
        return baseurl[:-1] + path
    if count == 1:
        return baseurl + path
    if count == 0:
        return baseurl + '/' + path
    raise ValueError("Invalid baseurl/path combination")


class ContentType:
    urlencoded = {'Content-Type': 'application/x-www-form-urlencoded'}
    textplain = {'Content-Type': 'text/plain'}
    application_json = {'Content-Type': 'application/json'}


class HttpClientBase():
    def __init__(self, session_id_key):
        self.timeout = config.http['timeout']
        self.retries = config.http['retries']
        self.ssl_error_handling = config.http['ssl']
        self.session = requests.Session()
        self.session.verify = self.ssl_error_handling != 'Trust'
        self._session_id_key = session_id_key

    def dispatch(self, ctera_request):
        attempt = 0
        while attempt < self.retries:
            try:
                return self._do_dispatch(ctera_request)
            except requests_exceptions.HTTPError as error:
                http_exception = HTTPException(error)
                if http_exception.response.code != 504:
                    raise http_exception
                logging.getLogger().warning('Server timed out. %s', {'attempt': (attempt + 1)})
            except requests_exceptions.Timeout:
                self.on_timeout(attempt)
            except requests_exceptions.SSLError as error:
                self.on_ssl_error(error.request)
                attempt = -1
            except requests_exceptions.ConnectionError as error:
                self._on_unreachable(error)
            except requests_exceptions.RequestException as error:
                logging.getLogger().warning(error)

            attempt = attempt + 1
            if attempt < self.retries:
                time.sleep(self.timeout)
        logging.getLogger().error('Reached maximum number of retries. %s', {'retries': self.retries, 'timeout': self.timeout})
        raise ExhaustedException(self.retries, self.timeout)

    def _do_dispatch(self, ctera_request):
        response = self.session.request(ctera_request.method, ctera_request.url, **ctera_request.kwargs)
        response.raise_for_status()
        return (response.request, response)

    @staticmethod
    def _on_unreachable(error):
        parsed_url = urllib.parse.urlparse(error.request.url)
        logging.getLogger().error('Cannot reach target host. %s', {'host': parsed_url.hostname, 'port': parsed_url.port})
        socket_error = Object()
        socket_error.message = str(error)
        raise HostUnreachable(socket_error, parsed_url.hostname, parsed_url.port, parsed_url.scheme.upper())

    @staticmethod
    def on_timeout(attempt):
        logging.getLogger().warning('Request timed out. %s', {'attempt': (attempt + 1)})

    def on_ssl_error(self, request):
        parsed_url = urllib.parse.urlparse(request.url)
        if self.should_trust(parsed_url.hostname, parsed_url.port):
            self.trust(parsed_url.hostname, parsed_url.port)
        else:
            raise SSLException(parsed_url.hostname, parsed_url.port, 'Cancelled by user')

    def should_trust(self, host, port):
        if self.ssl_error_handling == 'Consent':
            return ask('Proceed to ' + host + ':' + str(port) + '?')
        raise SSLException(host, port, 'Configuration file requires the use of trusted certificates')

    def trust(self, _host, _port):
        self.session.verify = False  # CertificateServices.save_cert_from_server(host, port)

    def get_session_id(self):
        return self.session.cookies.get(self._session_id_key)

    def set_session_id(self, session_id):
        self.session.cookies.set(self._session_id_key, session_id)

    def set_custom_headers(self, headers):
        """
        Add custom headers that will be included in every http request.

        :param dict headers: the headers, represented as a key-value str dict
        """
        self.session.headers.update(headers)


class HttpClientRequest():
    def __init__(self, method, url, **kwargs):
        self.method = method
        self.url = url
        self.kwargs = kwargs


class HttpClientRequestGet(HttpClientRequest):
    def __init__(self, url, params=None, headers=None, stream=None):
        super().__init__('GET', url, params=params, headers=headers, stream=stream)


class HttpClientRequestPost(HttpClientRequest):
    def __init__(self, url, headers=None, data=None):
        super().__init__('POST', url, headers=headers, data=data)


class HttpClientRequestPut(HttpClientRequest):
    def __init__(self, url, headers=None, data=None):
        super().__init__('PUT', url, headers=headers, data=data)


class HttpClientRequestDelete(HttpClientRequest):
    def __init__(self, url, headers=None):
        super().__init__('DELETE', url, headers=headers)


class HttpClientRequestMkcol(HttpClientRequest):
    def __init__(self, url, headers=None):
        super().__init__('MKCOL', url, headers=headers)


class HttpClientRequestCopyMove(HttpClientRequest):
    def __init__(self, method, src, dest, overwrite, headers=None):
        headers = merge({
            'Destination': dest,
            'Overwrite': 'T' if overwrite else 'F'
        }, headers)
        super().__init__(method, src, headers=headers)


class HttpClientRequestCopy(HttpClientRequestCopyMove):
    def __init__(self, src, dest, overwrite, headers=None):
        super().__init__('COPY', src, dest, overwrite, headers=headers)


class HttpClientRequestMove(HttpClientRequestCopyMove):
    def __init__(self, src, dest, overwrite, headers=None):
        super().__init__('MOVE', src, dest, overwrite, headers=headers)


class HTTPClient(HttpClientBase):

    def get(self, url, params=None, headers=None, stream=None):
        return self.dispatch(HttpClientRequestGet(url, params=params, headers=headers, stream=stream))

    def post(self, url, headers=None, data='', urlencode=False):
        if urlencode:
            data = urllib.parse.urlencode(data).encode('utf-8')
        return self.dispatch(HttpClientRequestPost(url, headers=headers, data=data))

    def put(self, url, headers=None, data=''):
        return self.dispatch(HttpClientRequestPut(url, headers=headers, data=data))

    def delete(self, url, headers=None):
        return self.dispatch(HttpClientRequestDelete(url, headers=headers))

    def mkcol(self, url, headers=None):
        return self.dispatch(HttpClientRequestMkcol(url, headers=headers))

    def copy(self, src, dest, overwrite, headers=None):
        return self.dispatch(HttpClientRequestCopy(src, dest, overwrite, headers=headers))

    def move(self, src, dest, overwrite, headers=None):
        return self.dispatch(HttpClientRequestMove(src, dest, overwrite, headers=headers))

    def multipart(self, url, form_data, monitor_function_generator=None):
        encoder = MultipartEncoder(form_data)
        if monitor_function_generator:
            encoder = MultipartEncoderMonitor(encoder, callback=monitor_function_generator(encoder.len))
        return self.dispatch(HttpClientRequestPost(url, headers={'Content-Type': encoder.content_type}, data=encoder))

    def upload(self, url, form_data):
        def upload_monitor_generator(total_length):
            def upload_monitor(encoder):
                logging.getLogger().info(
                    'Uploaded %(percent)s - (%(uploaded)s out of %(size)s)',
                    dict(
                        percent=str(round(encoder.bytes_read/total_length*100))+'%',
                        uploaded=encoder.bytes_read,
                        size=total_length
                    )
                )
            return upload_monitor
        return self.multipart(url, form_data, upload_monitor_generator)



class XMLTypes:
    OBJ = 'obj'
    ATT = 'att'
    VAL = 'val'
    LIST = 'list'
    ID = 'id'
    CLASS = 'class'
    UUID = 'uuid'
    DB = 'db'
    VERSION = 'version'
    FIRMWARE = 'firmware'
    NS = 'xmlns:xsi'
    LOCATION = 'xsi:noNamespaceSchemaLocation'



from ...common import Object
from . import ls
from ...exception import RemoteDirectoryNotFound


class SrcDstParam(Object):

    __instance = None

    @staticmethod
    def instance(src, dest=None):
        SrcDstParam(src, dest)
        return SrcDstParam.__instance

    def __init__(self, src, dest=None):
        self._classname = self.__class__.__name__
        self.src = src
        self.dest = dest
        SrcDstParam.__instance = self  # pylint: disable=unused-private-member


class ActionResourcesParam(Object):

    __instance = None

    @staticmethod
    def instance():
        ActionResourcesParam()
        return ActionResourcesParam.__instance

    def __init__(self):
        self._classname = self.__class__.__name__
        self.urls = []
        ActionResourcesParam.__instance = self  # pylint: disable=unused-private-member

    def add(self, param):
        self.urls.append(param)


class CreateShareParam(Object):

    __instance = None

    @staticmethod
    def instance(path, access, expire_on):
        CreateShareParam(path, access, expire_on)
        return CreateShareParam.__instance

    def __init__(self, path, access, expire_on):
        self._classname = self.__class__.__name__
        self.url = path
        self.share = Object()
        self.share._classname = 'ShareConfig'
        self.share.accessMode = access
        self.share.protectionLevel = 'publicLink'
        self.share.expiration = expire_on
        self.share.invitee = Object()
        self.share.invitee._classname = 'Collaborator'
        self.share.invitee.type = 'external'
        CreateShareParam.__instance = self  # pylint: disable=unused-private-member


def get_resource_info(ctera_host, path):
    response = ls.ls(ctera_host, path, depth=0)
    if response.root is None:
        raise RemoteDirectoryNotFound(path.fullpath())
    return response.root



from ..exception import CTERAException
from ..common import Object


class DirectoryTree:

    def __init__(self, tree):
        self.root = self._decorate(tree)

    def _decorate(self, root):
        root._parent = None  # pylint: disable=protected-access
        nodes = [root]
        while nodes:
            node = nodes.pop(0)
            if self._has_children(node):
                for child_node in node.children:
                    child_node._parent = node  # pylint: disable=protected-access
                    nodes.append(child_node)
        return root

    def include_folder(self, path):
        self._configure_selection(path, True, True)

    def include_file(self, path):
        self._configure_selection(path, False, True)

    def exclude_folder(self, path):
        self._configure_selection(path, True, False)

    def exclude_file(self, path):
        self._configure_selection(path, False, False)

    def _configure_selection(self, path, is_dir, include):
        self._check_root_dir(path)
        node, parts = self._lookup(path)
        if not parts:
            if is_dir and not DirectoryTree._is_dir(node):
                raise CTERAException('Expected to find a directory but found file', None, path=path)
            if DirectoryTree._is_dir(node) and not is_dir:
                raise CTERAException('Expected to find a file but found a directory', None, path=path)
            node.isIncluded = include
            node.children = None
        else:
            self._populate_selection(node, parts, is_dir, include)

    def remove_selection(self, path):
        self._check_root_dir(path)
        node, parts = self._lookup(path)
        if not parts:
            if node._parent is not None:  # pylint: disable=protected-access
                self._remove_child(node._parent, node.name)  # pylint: disable=protected-access
            else:
                node.isIncluded = False
                node.children = None
        else:
            raise CTERAException('Could not find directory path', None, path=path)

    def select_all(self):
        self.root.isIncluded = True
        self.root.children = None

    def unselect_all(self):
        self.root.isIncluded = False
        self.root.children = None

    def _check_root_dir(self, path):
        parts = path.split('/')
        if self.root.name != parts[0]:
            raise CTERAException('Invalid root directory', None, input=parts[0], should_start_with=self.root.name)

    def _lookup(self, path):
        parts = path.split('/')[1:]
        node = self.root
        while parts:
            if self._has_children(node):
                child_name = parts.pop(0)
                child_node = self._get_child(node, child_name)
                if child_node is None:
                    parts.insert(0, child_name)
                    break
                node = child_node
            else:
                break
        return (node, parts)

    @staticmethod
    def _is_dir(node):
        return getattr(node, '_classname', 'DirEntry') == 'DirEntry'

    # pylint: disable=R6301
    def _has_children(self, node):
        return node.children is not None

    def _get_child(self, node, child_name):
        if self._has_children(node):
            for child in node.children:
                if str(child.name) == child_name:
                    return child
        return None

    # pylint: disable=R6301
    def _add_child(self, parent, node):
        node._parent = parent  # pylint: disable=protected-access
        if parent.children is None:
            parent.children = []
        parent.children.append(node)

    def _remove_child(self, parent, child_name):
        ret = None
        if self._has_children(parent):
            for i, child in enumerate(parent.children):
                if child.name == child_name:
                    ret = parent.children.pop(i)
                    if not parent.children:
                        parent.children = None
                    break
        return ret

    def _populate_selection(self, parent, parts, is_dir, include):
        descendant_name = parts.pop()
        descendant = self._get_dir_entry(descendant_name, include) if is_dir else self._get_file_entry(descendant_name, include)
        for part in parts:
            child_node = self._get_dir_entry(part, parent.isIncluded)
            self._add_child(parent, child_node)
            parent = child_node
        self._add_child(parent, descendant)

    def _get_file_entry(self, name, include):
        return self._get_entry(False, name, include)

    def _get_dir_entry(self, name, include):
        return self._get_entry(True, name, include)

    # pylint: disable=R6301
    def _get_entry(self, is_dir, name, include):
        param = Object()
        param.displayName = None
        param.name = name
        param.isIncluded = include
        if is_dir:
            param._classname = 'DirEntry'  # pylint: disable=protected-access
            param.children = None
        else:
            param._classname = 'FileEntry'  # pylint: disable=protected-access
        return param



from .base_command import BaseCommand


class Reports(BaseCommand):
    """
    Reports APIs
    """
    def storage(self):
        """
        Retrieve the portals statistics report.\n
        To retrieve this report, you must first browse the Global Administration Portal, using: `GlobalAdmin.portals.browse_global_admin()`
        """
        return self._get_report('storageLocationsStatisticsReport')

    def portals(self):
        """
        Retrieve the storage statistics report.\n
        To retrieve this report, you must first browse the Global Administration Portal, using: `GlobalAdmin.portals.browse_global_admin()`
        """
        return self._get_report('portalsStatisticsReport')

    def folders(self):
        """
        Retrieve the cloud folders statistics report.\n
        To retrieve this report, you must first browse the Virtual Portal that contains the report, using: `GlobalAdmin.portals.browse()`
        """
        return self._get_report('foldersStatisticsReport')

    def folder_groups(self):
        """
        Retrieve the folder groups statistics report.\n
        To retrieve this report, you must first browse the Virtual Portal that contains the report, using: `GlobalAdmin.portals.browse()`
        """
        return self._get_report('folderGroupsStatisticsReport')

    def devices(self):
        """
        Retrieve the devices statistics report.\n
        To retrieve this report, you must first browse the Virtual Portal that contains the report, using: `GlobalAdmin.portals.browse()`
        """
        return self._get_report('devicesStatisticsReport')

    def _get_report(self, report_name):
        return self._portal.get(f'/reports/{report_name}')



from urllib.parse import urljoin

from .path import CTERAPath
from . import common
from ...common import Object
from ...lib import FileAccessBase
from ...exception import RemoteFileSystemException


class FileAccess(FileAccessBase):

    def _get_single_file_url(self, path):
        return path.fullpath()

    def _get_multi_file_url(self, cloud_directory, files):
        folder_uid = self._get_cloud_folder_uid(cloud_directory)
        return f'{self._ctera_host.context}/folders/folders/{folder_uid}'

    @property
    def _use_file_url_for_multi_file_url(self):
        return True

    def _get_multi_file_object(self, cloud_directory, files):
        files_obj = Object()
        files_obj.paths = ['/'.join([cloud_directory.fullpath(), filename]) for filename in files]
        files_obj.snapshot = None
        files_obj.password = None
        files_obj.portalName = None
        files_obj.showDeleted = False
        return files_obj

    def _get_upload_url(self, dest_path):
        folder_uid = self._get_cloud_folder_uid(dest_path)
        return f'{self._ctera_host.context}/upload/folders/{folder_uid}'

    def _get_upload_form(self, local_file_info, fd, dest_path):
        return dict(
            name=local_file_info['name'],
            Filename=local_file_info['name'],
            fullpath=urljoin(
                self._ctera_host.base_file_url,
                CTERAPath(local_file_info['name'], dest_path.fullpath()).encoded_fullpath()
            ),
            fileSize=local_file_info['size'],
            file=(local_file_info['name'], fd, local_file_info['mimetype'][0])
        )

    def _get_cloud_folder_uid(self, path):
        resource_info = common.get_resource_info(self._ctera_host, path)
        if not resource_info.isFolder:
            raise RemoteFileSystemException('The destination path is not a directory', None, path=path.fullpath())
        return resource_info.cloudFolderInfo.uid



import logging

from ..exception import CTERAException


def remote_access(Gateway, Portal):
    tenant = Portal.session().tenant()
    device = Gateway.host()
    logging.getLogger().info("Enabling remote access. %s", {'tenant': tenant, 'device': device})

    ticket = obtain_ticket(Portal, device)
    Gateway.session().enable_remote_access()
    login(Gateway, ticket)
    logging.getLogger().info("Enabled remote access. %s", {'tenant': tenant, 'device': device})


def login(Gateway, ticket):
    logging.getLogger().debug("Logging in using SSO ticket. %s", {'device': Gateway.host()})
    Gateway.get('/ssologin', {'ticket': ticket})


def obtain_ticket(Portal, device_name):
    tenant = Portal.session().tenant()
    url = f'/portals/{tenant}/devices/{device_name}'
    logging.getLogger().debug("Obtaining SSO ticket. %s", {'tenant': tenant, 'device': device_name})

    ticket = Portal.execute(url, 'singleSignOn')
    if not ticket:
        logging.getLogger().error('Could not obtain SSO ticket. %s', {'tenant': tenant, 'device': device_name})
        raise CTERAException('Could not obtain SSO ticket.')

    logging.getLogger().debug("Obtained SSO ticket. %s", {'tenant': tenant, 'device': device_name})

    return ticket



from pathlib import PurePosixPath
from urllib.parse import quote, unquote
import re

from ...exception import InputError
from ...common import Object


class CTERAPath:

    def __init__(self, item, basepath):
        if isinstance(item, str):
            self.basepath = PurePosixPath(basepath)
            self.relativepath = PurePosixPath(item)
        elif isinstance(item, Object) and hasattr(item, '_classname') and item._classname == 'ResourceInfo':
            href = unquote(item.href)
            match = re.search('^/(ServicesPortal|admin)/webdav', href)
            start, end = match.span()
            self.basepath = PurePosixPath(href[start: end])
            self.relativepath = PurePosixPath(href[end + 1:])
        else:
            raise InputError('Invalid path', item, 'comma separated str path segments, or a ResourceInfo object')

        if self.relativepath.root == '/' or self.basepath.joinpath(self.relativepath) == self.relativepath:
            raise InputError(
                'You must specify a relative path. Omit leading / characters',
                str(self.relativepath),
                re.sub(r'^/*', '', str(self.relativepath))
            )

    def name(self):
        return self.relativepath.name

    def parent(self):
        return CTERAPath(str(self.relativepath.parent), str(self.basepath))

    def fullpath(self):
        return str(self.basepath.joinpath(self.relativepath))

    def encoded_fullpath(self):
        return quote(self.fullpath())

    def encoded_parent(self):
        return quote(str(self.parent()))

    def joinpath(self, path):
        return CTERAPath(str(self.relativepath.joinpath(path)), str(self.basepath))

    def parts(self):
        return self.relativepath.parts

    def __str__(self):
        return self.fullpath()



import logging

from ..exception import InputError
from .base_command import BaseCommand
from .types import UserGroupEntry


class Groups(BaseCommand):

    def get(self, name=None):
        """
        Get Group. If a group name was not passed as an argument, a list of all local groups will be retrieved

        :param str,optional name: Name of the group
        """
        return self._gateway.get('/config/auth/groups' + ('' if name is None else ('/' + name)))

    def add_members(self, group, members):
        """
        Add members to a group

        :param str group: Name of the group
        :param list[cterasdk.edge.types.UserGroupEntry] members: List of users and groups to add to the group
        """
        Groups._validate_members(members)
        new_member_dict = {
            member.principal_type + '#' + member.name: member.to_server_object()
            for member in members
        }

        current_members = self._gateway.get('/config/auth/groups/' + group + '/members')
        for current_member in current_members:
            user_group_entry = UserGroupEntry.from_server_object(current_member)
            user_group_entry_key = user_group_entry.principal_type + '#' + user_group_entry.name
            if user_group_entry_key not in new_member_dict:
                new_member_dict[user_group_entry_key] = current_member

        members = [v for k, v in new_member_dict.items()]

        logging.getLogger().info('Adding group members. %s', {'group': group})

        self._gateway.put('/config/auth/groups/' + group + '/members', members)

        logging.getLogger().info('Group members added. %s', {'group': group})

    def remove_members(self, group, members):
        """
        Remove members from a group

        :param str group: Name of the group
        :param list[cterasdk.edge.types.UserGroupEntry] members: List of users and groups to remove from the group
        """
        Groups._validate_members(members)
        remove_members_dict = {
            member.principal_type + '#' + member.name: True
            for member in members
        }

        current_members = self._gateway.get('/config/auth/groups/' + group + '/members')
        members = []
        for member in current_members:
            user_group_entry = UserGroupEntry.from_server_object(member)
            if not remove_members_dict.get(user_group_entry.principal_type + '#' + user_group_entry.name, False):
                members.append(member)

        logging.getLogger().info('Removing group members. %s', {'group': group})

        self._gateway.put('/config/auth/groups/' + group + '/members', members)

        logging.getLogger().info('Group members removed. %s', {'group': group})

    @staticmethod
    def _validate_members(members):
        if not isinstance(members, list):
            raise InputError('Invalid members list format', repr(members), '[cterasdk.edge.types.UserGroupEntry, ...]')
        for member in members:
            if not isinstance(member, UserGroupEntry):
                raise InputError('Invalid access control entry format', repr(member), 'cterasdk.edge.types.UserGroupEntry')



import logging

from .common import SrcDstParam, ActionResourcesParam


def move(ctera_host, src, dest):
    return move_multi(ctera_host, [src], dest)


def move_multi(ctera_host, src, dest):
    move_param = ActionResourcesParam.instance()

    for path in src:
        logging.getLogger().info('Moving item. %s', {'path': str(path.relativepath), 'to': str(dest.relativepath)})
        move_param.add(SrcDstParam.instance(src=path.fullpath(), dest=dest.joinpath(path.name()).fullpath()))

    return ctera_host.execute('', 'moveResources', move_param)



import logging

from ..exception import InputError
from .enum import License
from .base_command import BaseCommand


class Licenses(BaseCommand):
    """ Edge Filer License Configuration APIs """

    def __init__(self, gateway):
        super().__init__(gateway)
        self.local = LocalLicenses(self._gateway)

    @staticmethod
    def infer(ctera_license):
        if License.__dict__.get(ctera_license) is None:
            logging.getLogger().error('Invalid license type. %s', {'license': ctera_license})
            options = [v for k, v in License.__dict__.items() if not k.startswith('_')]
            raise InputError('Invalid license type', ctera_license, options)

        ctera_license = License.__dict__.get(ctera_license)
        return 'vGateway' + ('' if ctera_license == 'EV16' else ctera_license[2:])

    def apply(self, ctera_license):
        """
        Apply a license

        :param cterasdk.edge.enum.License ctera_license: License type
        """
        inferred_license = Licenses.infer(ctera_license)

        logging.getLogger().info('Applying license. %s', {'license': ctera_license})

        self._gateway.put('/config/device/activeLicenseType', inferred_license)

        logging.getLogger().info('License applied. %s', {'license': ctera_license})

    def get(self):
        """
        Get the current Gateway License
        """
        inferred_license = self._gateway.get('/config/device/activeLicenseType')
        if inferred_license == 'NA':
            return inferred_license
        if len(inferred_license) == 8:
            inferred_license = inferred_license + '16'
        return 'EV' + inferred_license[8:]


class LocalLicenses(BaseCommand):
    """ Edge Filer Local License Configuration APIs """

    def get(self):
        """
        Retrieve a list of local licenses installed
        """
        return self._gateway.get('/config/device/licenses')

    def add(self, code):
        """
        Install a local license. Use this option when running the Edge Filer as a standalone appliance.

        :param str code: License code
        """
        logging.getLogger().info('Installing license. %s', {'code': code})
        response = self._gateway.add('/config/device/licenses', code)
        logging.getLogger().info("License added. %s", {'code': code})
        return response

    def clear(self):
        """
        Remove all local licenses
        """
        self._gateway.put('/config/device/licenses', [])



from pathlib import PurePosixPath
import re
from urllib.parse import quote

from ...exception import InputError


class CTERAPath:

    def __init__(self, relativepath, basepath):
        self.basepath = PurePosixPath(basepath)
        self.relativepath = PurePosixPath(relativepath)
        if self.basepath.joinpath(self.relativepath) == self.relativepath:
            raise InputError(
                'You must specify a relative path. Omit leading / characters',
                str(self.relativepath),
                re.sub(r'^/*', '', str(self.relativepath))
            )

    def name(self):
        return self.relativepath.name

    def parent(self):
        return CTERAPath(str(self.relativepath.parent), str(self.basepath))

    def fullpath(self):
        return str(self.basepath.joinpath(self.relativepath))

    def encoded_fullpath(self):
        return quote(self.fullpath())

    def encoded_parent(self):
        return quote(str(self.parent()))

    def joinpath(self, path):
        return CTERAPath(str(self.relativepath.joinpath(path)), str(self.basepath))

    def parts(self):
        return self.relativepath.parts

    def __str__(self):
        return self.fullpath()



from .http import HTTPClient, ContentType, HTTPException, HTTPResponse, geturi
from ..convert import fromxmlstr, fromjsonstr, toxmlstr, tojsonstr
from ..exception import CTERAClientException
from ..lib import Command
from ..common import Object
from ..transcript import transcribe
from .. import config


class CTERAClient:

    def __init__(self, session_id_key):
        self.http_client = HTTPClient(session_id_key)

    def get(self, baseurl, path, params=None):
        function = Command(HTTPClient.get, self.http_client, geturi(baseurl, path), params if params else {})
        return self._execute(function)

    def download(self, baseurl, path, params):
        function = Command(HTTPClient.get, self.http_client, geturi(baseurl, path), params, None, True)
        return self._execute(function, return_function=CTERAClient.file_descriptor)

    def download_zip(self, baseurl, path, form_data):
        function = Command(HTTPClient.post, self.http_client, geturi(baseurl, path), ContentType.urlencoded, form_data, True)
        return self._execute(function, return_function=CTERAClient.file_descriptor)

    def get_multi(self, baseurl, path, paths):
        return self.db(baseurl, path, "get-multi", paths)

    def put(self, baseurl, path, data):
        function = Command(HTTPClient.put, self.http_client, geturi(baseurl, path), ContentType.textplain, toxmlstr(data))
        return self._execute(function)

    def post(self, baseurl, path, data):
        function = Command(HTTPClient.post, self.http_client, geturi(baseurl, path), ContentType.textplain, toxmlstr(data))
        return self._execute(function)

    def form_data(self, baseurl, path, form_data):
        function = Command(HTTPClient.post, self.http_client, geturi(baseurl, path), ContentType.urlencoded, form_data, True)
        return self._execute(function)

    def execute(self, baseurl, path, name, param=None):
        return self._ctera_exec(baseurl, path, 'user-defined', name, param)

    def delete(self, baseurl, path):
        function = Command(HTTPClient.delete, self.http_client, geturi(baseurl, path))
        return self._execute(function)

    def mkcol(self, baseurl, path):
        function = Command(HTTPClient.mkcol, self.http_client, geturi(baseurl, path))
        return self._execute(function)

    def copy(self, baseurl, src, dest, overwrite):
        function = Command(HTTPClient.copy, self.http_client, geturi(baseurl, src), geturi(baseurl, dest), overwrite)
        return self._execute(function)

    def move(self, baseurl, src, dest, overwrite):
        function = Command(HTTPClient.move, self.http_client, geturi(baseurl, src), geturi(baseurl, dest), overwrite)
        return self._execute(function)

    def db(self, baseurl, path, name, param):
        return self._ctera_exec(baseurl, path, 'db', name, param)

    def multipart(self, baseurl, path, form_data):
        function = Command(HTTPClient.multipart, self.http_client, geturi(baseurl, path), form_data)
        return self._execute(function)

    def upload(self, baseurl, path, form_data):
        function = Command(HTTPClient.upload, self.http_client, geturi(baseurl, path), form_data)
        return self._execute(function)

    def _ctera_exec(self, baseurl, path, exec_type, name, param):
        obj = Object()
        obj.type = exec_type
        obj.name = name
        obj.param = param
        function = Command(HTTPClient.post, self.http_client, geturi(baseurl, path), ContentType.textplain, toxmlstr(obj))
        return self._execute(function)

    def get_session_id(self):
        return self.http_client.get_session_id()

    def set_session_id(self, session_id):
        self.http_client.set_session_id(session_id)

    def set_authorization_headers(self, headers):
        self.http_client.set_custom_headers(headers)

    @staticmethod
    def fromxmlstr(request, response):
        response = transcribe_request(request, response, True)
        return fromxmlstr(response.text)

    @staticmethod
    def file_descriptor(request, response):
        return transcribe_request(request, response, False)

    @staticmethod
    def _execute(function, return_function=None):
        return_function = return_function or CTERAClient.fromxmlstr
        return execute_request(function, return_function)


def execute_request(request_function, return_function):
    """
    Execute an HTTP request
    """
    try:
        request, response = request_function()
        return return_function(request, response)
    except HTTPException as http_error:
        client_error = CTERAClientException()
        client_error.__dict__ = http_error.__dict__.copy()
        raise client_error


def transcribe_request(request, response, transcribe_response=True):
    """
    Transcribe the HTTP request
    """
    if not config.transcript['disabled']:
        if transcribe_response:
            response = HTTPResponse(response)
            transcribe.transcribe(request, response)
        else:
            transcribe.transcribe(request)
    return response


class RESTClient:

    def __init__(self, http_client=None, session_id_key=None):
        self.http_client = http_client if http_client else HTTPClient(session_id_key)

    def get(self, baseurl, path, params=None):
        function = Command(HTTPClient.get, self.http_client, geturi(baseurl, path), params if params else {})
        return self._execute(function)

    def put(self, baseurl, path, data):
        function = Command(HTTPClient.put, self.http_client, geturi(baseurl, path), ContentType.application_json, tojsonstr(data))
        return self._execute(function)

    def post(self, baseurl, path, data):
        function = Command(HTTPClient.post, self.http_client, geturi(baseurl, path), ContentType.application_json, tojsonstr(data))
        return self._execute(function)

    def delete(self, baseurl, path):
        function = Command(HTTPClient.delete, self.http_client, geturi(baseurl, path))
        return self._execute(function)

    @staticmethod
    def fromjsonstr(request, response):
        response = transcribe_request(request, response, True)
        return fromjsonstr(response.text)

    @staticmethod
    def _execute(function):
        return execute_request(function, RESTClient.fromjsonstr)


class MigrationClient(RESTClient):

    XSRF_TOKEN_ID = 'x-mt-x'

    def login(self, baseurl, path):
        _, response = self.http_client.get(geturi(baseurl, path))
        xsrf_token = response.headers.get(MigrationClient.XSRF_TOKEN_ID, None)
        self.http_client.set_custom_headers({MigrationClient.XSRF_TOKEN_ID: xsrf_token})



from .common import Object
from .convert import tojsonstr


class CTERAException(Exception):

    def __init__(self, message=None, instance=None, **kwargs):
        super().__init__()
        self.classname = self.__class__.__name__
        self.join(instance)
        if message is not None:
            self.message = message
        self.put(**kwargs)

    def put(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def join(self, instance=None):
        if instance is not None:
            self.__dict__ = instance.__dict__.copy()

    def __str__(self):
        return tojsonstr(self)


class CTERAClientException(CTERAException):
    pass


class ObjectNotFoundException(CTERAException):

    def __init__(self, message, object_ref, **kwargs):
        super().__init__(message, None, **kwargs)
        self.response = Object()
        self.response.code = 404
        self.response.reason = 'Not Found'
        self.response.body = Object()
        self.response.body.rc = 1
        self.response.body.msg = f"Object '{object_ref}' not found"


class CTERAConnectionError(CTERAException):

    def __init__(self, message, instance, host, port, protocol, **kwargs):
        super().__init__(message, instance, host=host, port=port, protocol=protocol, **kwargs)


class ConnectionTimeout(CTERAException):

    def __init__(self, message, seconds, **kwargs):
        super().__init__(message, None, seconds=seconds, **kwargs)


class HostUnreachable(CTERAConnectionError):

    def __init__(self, instance, host, port, protocol):
        super().__init__("Unable to reach host", instance, host=host, port=port, protocol=protocol)


class ExhaustedException(ConnectionTimeout):

    def __init__(self, retries, timeout):
        super().__init__("Don't blame me for lack of trying", retries * timeout, retries=retries, timeout=timeout)


class PythonVersionException(CTERAException):

    def __init__(self, version):
        super().__init__('You are running an unsupported version of Python', None, version=version)


class SSLException(CTERAConnectionError):

    def __init__(self, host, port, reason):
        super().__init__('Untrusted security certificate', None, host=host, port=port, protocol='TLS', reason=reason)


class InputError(CTERAException):

    def __init__(self, message, expression, options):
        CTERAException.__init__(self, message, None, expression=expression, options=options)


class ConsentException(CTERAException):

    def __init__(self):
        CTERAException.__init__(self, 'Every decision is liberating, even if it leads to disaster')


class FileSystemException(CTERAException):
    pass


class RenameException(FileSystemException):
    def __init__(self, dirpath, src, dst):
        FileSystemException.__init__(self, 'Could not rename file', None, dirpath=dirpath, src=src, dst=dst)


class LocalDirectoryNotFound(FileSystemException):

    def __init__(self, path):
        FileSystemException.__init__(self, 'Could not find local directory', None, path=path)


class LocalFileNotFound(FileSystemException):

    def __init__(self, path):
        FileSystemException.__init__(self, 'Could not find local file', None, path=path)


class LocalPathNotFound(FileSystemException):

    def __init__(self, path):
        FileSystemException.__init__(self, 'Path does not exist', None, path=path)


class RemoteFileSystemException(CTERAException):
    pass


class RemoteDirectoryNotFound(RemoteFileSystemException):

    def __init__(self, path):
        super().__init__('Could not find remote directory', None, path=path)



import logging
import time

from ..exception import CTERAException, CTERAClientException, HostUnreachable, ExhaustedException
from .base_command import BaseCommand


class Startup(BaseCommand):
    """
    Server Startup APIs
    """

    Started = 'Started'

    def status(self):
        """
        Get the server startup status
        """
        response = None
        try:
            response = self._portal.get(f'/{self._portal.context}/startup', use_file_url=True)
        except CTERAClientException as error:
            return error.response.body.status
        return response.status

    def wait(self, retries=120, seconds=5):
        """
        Wait for server startup
        """
        attempt = 0
        while True:
            try:
                current_status = self.status()
                if current_status == Startup.Started:
                    logging.getLogger().info('Server started.')
                    break
                logging.getLogger().debug('Current server status. %s', {'status': current_status})
                attempt = attempt + 1
                if attempt >= retries:
                    logging.getLogger().error('Timed out. Server did not start in a timely manner.')
                    raise CTERAException('Timed out. Server did not start in a timely manner')
                time.sleep(seconds)
            except (HostUnreachable, ExhaustedException) as e:
                logging.getLogger().debug('Exception. %s', e.__dict__)
                attempt = attempt + 1
                time.sleep(seconds)



import logging

from .common import SrcDstParam, ActionResourcesParam


def copy(ctera_host, src, dest):
    return copy_multi(ctera_host, [src], dest)


def copy_multi(ctera_host, src, dest):
    copy_param = ActionResourcesParam.instance()

    for path in src:
        logging.getLogger().info('Copying item. %s', {'path': str(path.relativepath), 'to': str(dest.relativepath)})

        param = SrcDstParam.instance(src=path.fullpath(), dest=dest.joinpath(path.name()).fullpath())
        copy_param.add(param)

    return ctera_host.execute('', 'copyResources', copy_param)



from ..client import CTERAHost, authenticated
from ..core import connection, messaging
from ..core import activation
from ..core import antivirus
from ..core import buckets
from ..core import cli
from ..core import decorator
from ..core import directoryservice
from ..core import firmwares
from ..core import kms
from ..core import licenses
from ..core import login
from ..core import query
from ..core import logs
from ..core import portals
from ..core import plans
from ..core import reports
from ..core import servers
from ..core import devices
from ..core import credentials
from ..core import session
from ..core import storage_classes
from ..core import domains
from ..core import users
from ..core import groups
from ..core import cloudfs
from ..core import roles
from ..core import settings
from ..core import setup
from ..core import ssl
from ..core import startup
from ..core import syslog
from ..core import taskmgr
from ..core import templates
from ..core import uri
from ..core import files
from ..core import admins


class Portal(CTERAHost):  # pylint: disable=too-many-instance-attributes
    """
    Parent class for communicating with the Portal through either GlobalAdmin or ServicesPortal

    :ivar cterasdk.core.domains.Domains domains: Object holding the Portal domain APIs
    :ivar cterasdk.core.users.Users users: Object holding the Portal user APIs
    :ivar cterasdk.core.groups.Groups groups: Object holding the Portal group APIs
    :ivar cterasdk.core.admins.Administrators admins: Object holding the Portal GlobalAdmin users APIs
    :ivar cterasdk.core.plans.Plans plans: Object holding the Plan APIs
    :ivar cterasdk.core.reports.Reports reports: Object holding the Portal reports APIs
    :ivar cterasdk.core.devices.Devices devices: Object holding the Portal devices APIs
    :ivar cterasdk.core.directoryservice.DirectoryService directoryservice: Object holding the Directory Services APIs
    :ivar cterasdk.core.activation.Activation activation: Object holding the Portal activation APIs
    :ivar cterasdk.core.logs.Logs logs: Object holding the Portal logs APIs
    :ivar cterasdk.core.cloudfs.CloudFS cloudfs: Object holding the Portal CloudFS APIs
    :ivar cterasdk.core.roles.Roles roles: Object holding the Portal Role Settings APIs
    :ivar cterasdk.core.settings.Settings settings: Object holding the Portal Settings APIs
    :ivar cterasdk.core.settings.StorageClasses storage_classes: Object holding the Portal Storage Classes APIs
    :ivar cterasdk.core.taskmgr.Tasks tasks: Object holding the Portal Background Tasks APIs
    :ivar cterasdk.core.templates.Templates templates: Object holding the Portal Configuration Templates APIs
    :ivar cterasdk.core.firmwares.Firmwares firmwares: Object holding the Portal Firmware Repository APIs
    :ivar cterasdk.core.files.browser.FileBrowser files: Object holding the Portal File Browsing APIs
    """

    def __init__(self, host, port, https):
        """
        :param str host: The fully qualified domain name, hostname or an IPv4 address of the Gateway
        :param int port: Set a custom port number (0 - 65535)
        :param bool https: Set to True to require HTTPS
        """
        super().__init__(host, port, https)
        self._session = session.Session(self.host(), self.context)
        self.domains = domains.Domains(self)
        self.users = users.Users(self)
        self.groups = groups.Groups(self)
        self.admins = admins.Administrators(self)
        self.reports = reports.Reports(self)
        self.plans = plans.Plans(self)
        self.devices = devices.Devices(self)
        self.directoryservice = directoryservice.DirectoryService(self)
        self.cloudfs = cloudfs.CloudFS(self)
        self.activation = activation.Activation(self)
        self.files = files.CloudDrive(self, self.cloud_drive_base_path)
        self.backups = files.Backups(self, self.backups_base_path)
        self.logs = logs.Logs(self)
        self.roles = roles.Roles(self)
        self.settings = settings.Settings(self)
        self.storage_classes = storage_classes.StorageClasses(self)
        self.credentials = credentials.Credentials(self)
        self.tasks = taskmgr.Tasks(self)
        self.templates = templates.Templates(self)
        self.firmwares = firmwares.Firmwares(self)

    @property
    def base_api_url(self):
        return uri.api(self)

    @property
    def base_portal_url(self):
        return self.baseurl() + '/' + self.context

    @property
    def base_file_url(self):
        return self.baseurl()

    @property
    def _session_id_key(self):
        return 'JSESSIONID'

    @property
    def context(self):
        raise NotImplementedError("Implementing class must implement the context property")

    @property
    def cloud_drive_base_path(self):
        raise NotImplementedError("Implementing class must implement the cloud_drive_base_path property")

    @property
    def backups_base_path(self):
        raise NotImplementedError("Implementing class must implement the backups_base_path property")

    @property
    def _omit_fields(self):
        return super()._omit_fields + [
            'admins',
            'domains',
            'users',
            'groups',
            'reports',
            'plans',
            'devices',
            'directoryservice',
            'cloudfs',
            'activation',
            'files',
            'logs',
            'roles',
            'settings',
            'credentials',
            'tasks',
            'templates',
            'firmwares'
        ]

    @property
    def _login_object(self):
        return login.Login(self)

    def _is_authenticated(self, function, *args, **kwargs):
        def is_public(path):
            return path.startswith(f'/{self.context}/public')

        def is_setup(path):
            return path.startswith(f'/{self.context}/setup')

        def is_startup(path):
            return path.startswith(f'/{self.context}/startup')
        current_session = self.session()
        return current_session.authenticated() or current_session.initializing() or \
            is_public(args[0]) or is_setup(args[0]) or is_startup(args[0]) or \
            current_session.is_local_auth()

    def test(self):
        """ Verification check to ensure the target host is a Portal. """
        connection.test(self)
        return self.public_info()

    def public_info(self):
        """ Obtain the Portal's public info. """
        return self.get('/' + self.context + '/public/publicInfo', params={}, use_file_url=True)

    @decorator.update_current_tenant
    def put(self, path, value, use_file_url=False):
        return super().put(path, value, use_file_url=use_file_url)

    @authenticated
    def query(self, path, name, param):
        return query.query(self, path, name, param)

    @authenticated
    def show_query(self, path, name, param):
        return query.show(self, path, name, param)

    def iterator(self, path, param):
        return query.iterator(self, path, param)


class GlobalAdmin(Portal):  # pylint: disable=too-many-instance-attributes
    """
    Main class for Global Admin operations on a Portal

    :ivar cterasdk.core.portals.Portals portals: Object holding the Portals Management APIs
    :ivar cterasdk.core.licenses.Licenses licenses: Object holding the Portal license APIs
    :ivar cterasdk.core.cli.CLI cli: Object holding the Portal GlobalAdmin CLI APIs
    :ivar cterasdk.core.servers.Servers servers: Object holding the Servers Management APIs
    :ivar cterasdk.core.setup.Setup setup: Object holding the Portal setup APIs
    :ivar cterasdk.core.ssl.SSL ssl: Object holding the Portal SSL Certificate APIs
    :ivar cterasdk.core.kms.KMS kms: Object holding the Portal External Key Management APIs
    :ivar cterasdk.core.startup.Startup startup: Object holding the Portal startup APIs
    :ivar cterasdk.core.syslog.Syslog syslog: Object holding the Portal syslog APIs
    :ivar cterasdk.core.antivirus.Antivirus antivirus: Object holding the Portal Antivirus APIs
    :ivar cterasdk.core.buckets.Buckets buckets: Object holding the Portal Storage Node APIs
    :ivar cterasdk.core.messaging.Messaging messaging: Object holding the Portal Messaging Service Management APIs

    """

    def __init__(self, host, port=None, https=True):
        """
        :param str host: The fully qualified domain name, hostname or an IPv4 address of the Portal
        :param int,optional port: Set a custom port number (0 - 65535), If not set defaults to 80 for http and 443 for https
        :param bool,optional https: Set to True to require HTTPS, defaults to True
        """
        super().__init__(host, port, https)
        self.portals = portals.Portals(self)
        self.servers = servers.Servers(self)
        self.cli = cli.CLI(self)
        self.licenses = licenses.Licenses(self)
        self.kms = kms.KMS(self)
        self.setup = setup.Setup(self)
        self.ssl = ssl.SSL(self)
        self.startup = startup.Startup(self)
        self.syslog = syslog.Syslog(self)
        self.antivirus = antivirus.Antivirus(self)
        self.buckets = buckets.Buckets(self)
        self.messaging = messaging.Messaging(self)

    @property
    def _omit_fields(self):
        return super()._omit_fields + [
            'portals',
            'cli',
            'licenses',
            'kms',
            'servers',
            'setup',
            'ssl',
            'startup',
            'syslog',
            'antivirus',
            'buckets',
            'messaging'
        ]

    @property
    def context(self):
        return 'admin'

    @property
    def cloud_drive_base_path(self):
        return '/admin/webdav/Users'

    @property
    def backups_base_path(self):
        return '/admin/webdav/backupFolders'


class ServicesPortal(Portal):
    """
    Main class for Service operations on a Portal
    """

    def __init__(self, host, port=None, https=True):
        """
        :param str host: The fully qualified domain name, hostname or an IPv4 address of the Portal
        :param int,optional port: Set a custom port number (0 - 65535), If not set defaults to 80 for http and 443 for https
        :param bool,optional https: Set to True to require HTTPS, defaults to True
        """
        super().__init__(host, port, https)

    @property
    def context(self):
        return 'ServicesPortal'

    @property
    def cloud_drive_base_path(self):
        return '/ServicesPortal/webdav'

    @property
    def backups_base_path(self):
        return '/ServicesPortal/webdav/backups'



import logging

from .enum import Mode
from .base_command import BaseCommand


class AFP(BaseCommand):
    """ Gateway AFP APIs """

    def is_disabled(self):
        """ Check if the AFP server is disabled """
        return self._gateway.get('/config/fileservices/afp/mode') == Mode.Disabled

    def disable(self):
        """
        Disable AFP
        """
        logging.getLogger().info('Disabling AFP server.')

        self._gateway.put('/config/fileservices/afp/mode', Mode.Disabled)

        logging.getLogger().info('AFP server disabled.')



from .enum import DeviceType
from ..object.Gateway import Gateway
from ..object.Agent import Agent


def remote_command(Portal, device):
    port = Portal.port()
    https = Portal.https()

    ManagedDevice = None
    if device.deviceType in DeviceType.Gateways:
        ManagedDevice = Gateway(host=device.name, port=port, https=https, Portal=Portal)
    elif device.deviceType in DeviceType.Agents:
        ManagedDevice = Agent(host=device.name, port=port, https=https, Portal=Portal)
    elif device.deviceType == "Mobile":
        return device
    else:
        return device

    ManagedDevice.__dict__.update(device.__dict__.copy())

    return ManagedDevice
